@@ -6,6 +6,10 @@ import time
 from contextlib import closing
 import urllib.request
 import urllib.parse
+import urllib.error
+
+
+GP_JOB_TAG = 'GenePattern Python Client'
 
 
 class GPServer(object):
@@ -32,6 +36,24 @@ class GPServer(object):
         requests sent to GenePattern.
         """
         return 'Basic %s' % base64.b64encode(bytes(self.username + ':' + self.password, 'ascii')).decode('ascii')
+    
+    def system_message(self):
+        url = f"{self.url}/rest/v1/config/system-message"
+        request = urllib.request.Request(url)
+        response = urllib.request.urlopen(request)
+        return response.read().decode('utf-8')
+    
+    def login(self):
+        """Log in to the OAuth2 endpoint"""
+        safe_username = urllib.parse.quote(self.username)
+        safe_password = urllib.parse.quote(self.password)
+        url = f"{self.url}/rest/v1/oauth2/token?grant_type=password&username={safe_username}&password={safe_password}&client_id=GenePatternNotebook-{safe_username}"
+
+        request = urllib.request.Request(url)
+        response = urllib.request.urlopen(request, b'')
+        if response.getcode() != 200:
+            raise urllib.error.HTTPError(url, response.getcode(), 'Invalid username or password', response.getheaders(), None)
+        return json.loads(response.read())['access_token']
 
     def upload_file(self, file_name, file_path):
         """
@@ -88,7 +110,7 @@ class GPServer(object):
 
         # names should be a list of names,
         # values should be a list of **lists** of values
-        json_string = json.dumps({'lsid': job_spec.lsid, 'params': job_spec.params, 'tags': ['GenePattern Python Client']}, cls=GPJSONEncoder)
+        json_string = json.dumps({'lsid': job_spec.lsid, 'params': job_spec.params, 'tags': [GP_JOB_TAG]}, cls=GPJSONEncoder)
         if sys.version_info.major == 3:  # Handle conversion to bytes for Python 3
             json_string = bytes(json_string, 'utf-8')
         request = urllib.request.Request(self.url + '/rest/v1/jobs')
@@ -333,7 +355,7 @@ class GPJob(GPResource):
             self.get_info()
 
         # Lazily load children
-        if self.children is not None:
+        if self.children:
             return self.children
         else:
             if 'children' in self.info:
@@ -474,6 +496,27 @@ class GPJob(GPResource):
         Returns the URL of the job's status page on the GenePattern server
         """
         return self.server_data.url + "/pages/index.jsf?jobid=" + self.uri
+    
+    def get_permissions(self):
+        """Get the permissions object for the GP job"""
+        url = f'{self.server_data.url}/rest/v1/jobs/{self.job_number}/permissions'
+        request = urllib.request.Request(url)
+        if self.server_data.authorization_header() is not None:
+            request.add_header('Authorization', self.server_data.authorization_header())
+        request.add_header('User-Agent', 'GenePatternRest')
+
+        response = urllib.request.urlopen(request)
+        return json.loads(response.read())
+        
+    def set_permissions(self, permissions):
+        """Set the group permissions for the job"""
+        url = f'{self.server_data.url}/rest/v1/jobs/{self.job_number}/permissions'
+        data = json.dumps(permissions).encode('utf8')
+        request = urllib.request.Request(url, data=data, method='PUT')
+        if job.server_data.authorization_header() is not None:
+            request.add_header('Authorization', self.server_data.authorization_header())
+        request.add_header('User-Agent', 'GenePatternRest')
+        urllib.request.urlopen(request)
 
 
 class GPJobSpec(object):
@@ -656,7 +699,7 @@ class GPTaskParam(object):
         Returns whether the parameter is optional or required
         :return: Return True if optional, False if required
         """
-        if (('optional' in self.attributes and bool(self.attributes['optional'].strip())) and
+        if (('optional' in self.attributes and bool(self.attributes['optional'].strip())) or
                 ('minValue' in self.attributes and self.attributes['minValue'] == 0)):
             return True
         else:
@@ -704,8 +747,9 @@ class GPTaskParam(object):
         :return: Return True if the parameter allows multiple values, otherwise False
         """
         # note that maxValue means "max number of values", and is an integer, not a string
-        if (('maxValue' in self.attributes) and
-                (self.attributes['maxValue'] > 1)):
+        if ('maxValue' in self.attributes) and (self.attributes['maxValue'] > 1):
+            return True
+        elif ('numValues' in self.attributes) and ('+' in self.attributes['numValues']):
             return True
         else:
             return False
